package com.itko.lisa.ext.trilead;

import java.io.IOException;
import java.io.PrintWriter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Element;

import com.itko.lisa.test.TestCase;
import com.itko.lisa.test.TestDefException;
import com.itko.lisa.test.TestEvent;
import com.itko.lisa.test.TestExec;
import com.itko.lisa.test.TestNode;
import com.itko.lisa.test.TestRunException;
import com.itko.util.CloneImplemented;
import com.itko.util.StrUtil;
import com.itko.util.XMLUtils;

public class SFTPStep extends TestNode implements CloneImplemented {

	private static final long serialVersionUID = -2102844872961425683L;
	
	private static Log log = LogFactory.getLog("com.itko.lisa.ext.ssh");

	public static final String CMD_LS = "ls";
	public static final String CMD_READ = "read";
	public static final String CMD_OVERWRITE = "overwrite";
	public static final String CMD_FORCEWRITE = "forcewrite";
	public static final String CMD_CREATE = "create";
	public static final String CMD_DELETE = "delete";
	public static final String CMD_GET = "get";
	public static final String CMD_PUT = "put";
	public static final String CMD_MKDIR = "mkdir";
	public static final String CMD_CD = "cd";
	public static final String CMD_SITE = "site";
	public static final String CMD_QUIT = "quit";
	
	public static final String PROTOCOL_SFTP = "SFTP";
	public static final String PROTOCOL_SCP = "SCP";
	public static final String PROTOCOL_FTP = "FTP";
	
	public static final String TAG_HOSTNAME = "hostname";
	public static final String TAG_PORT = "port";
	public static final String TAG_USERNAME = "username";
	public static final String TAG_PASSWORD = "password";
	public static final String TAG_COMMAND = "command";
	public static final String TAG_PATH = "path";
	public static final String TAG_FILENAME = "filename";
	public static final String TAG_CONTENT = "content";
	public static final String TAG_PROTOCOL = "protocol";
	public static final String TAG_LOCAL_FILENAME = "localFilename";
	public static final String TAG_SAVE_TO_FILE = "saveToFile";
	public static final String TAG_AUTH = "auth";
	public static final String TAG_PRIVATE_KEY = "privateKey";
	public static final String TAG_PRIVATE_KEY_PASS = "privateKeyPass";
	
	
//	private static Map<String,GenericObjectPool> poolMap = new HashMap<String,GenericObjectPool>();
	protected static SFTPEngine sftpEngine = new SFTPEngine();
	protected static SCPEngine scpEngine = new SCPEngine();
	protected static FTPEngine ftpEngine = new FTPEngine();
	
	protected SFTPStepSettings settings;
	
	public void setSettings(SFTPStepSettings settings) {
		this.settings = settings;
	}
	
	public SFTPStepSettings getSettings() {
		return settings;
	}
	
	public static SFTPEngine getSFTPEngine() {
		return sftpEngine;
	}
	
	public static SCPEngine getSCPEngine() {
		return scpEngine;
	}
	
	public static FTPEngine getFTPEngine() {
		return ftpEngine;
	}	
	
	@Override
	protected void execute(TestExec ts) throws TestRunException {
		try {
			// Determine the proper engine to use based on the protocol
			String msg = "Unknown protocol.";
			if (settings.getProtocol().equalsIgnoreCase(PROTOCOL_SFTP)) {
				msg = sftpEngine.execute(settings, ts);
			} else if (settings.getProtocol().equalsIgnoreCase(PROTOCOL_SCP)) {
				msg = scpEngine.execute(settings, ts);
			} else if (settings.getProtocol().equalsIgnoreCase(PROTOCOL_FTP)) {
				msg = ftpEngine.execute(settings, ts);
			}
			ts.setLastResponse(msg);
			ts.raiseEvent(TestEvent.EVENT_NODERESPONSE, ts.getCurrentNode().getName(), msg);
			
		} catch (Exception e) {
			log.error(e.getMessage(), e);
			throw new TestRunException(e.getMessage(), e);
		}
	}

	@Override
	public void initialize(TestCase test, Element node) throws TestDefException {
		
		// Get the parameters
		settings = new SFTPStepSettings();
		settings.setHostname(XMLUtils.findChildGetItsText(node, TAG_HOSTNAME));
		settings.setPort(XMLUtils.findChildGetItsText(node, TAG_PORT));
		settings.setUsername(XMLUtils.findChildGetItsText(node, TAG_USERNAME));
		
		try {
			String pass = XMLUtils.findChildGetItsText(node, TAG_PASSWORD);
			settings.setPassword(StrUtil.decodeString(pass));
		} catch (IOException ioe) {
			settings.setPassword("");
		}
		
		// Decode the private key password
		try {
			String pass2 = XMLUtils.findChildGetItsText(node, TAG_PRIVATE_KEY_PASS);
			settings.setPrivateKeyPass(StrUtil.decodeString(pass2));
		} catch (IOException ioe) {
			settings.setPrivateKeyPass("");
		}

		
		settings.setProtocol(XMLUtils.findChildGetItsText(node, TAG_PROTOCOL));
		settings.setCommand(XMLUtils.findChildGetItsText(node, TAG_COMMAND));
		settings.setPath(XMLUtils.findChildGetItsText(node, TAG_PATH));
		settings.setFilename(XMLUtils.findChildGetItsText(node, TAG_FILENAME));
		settings.setContent(XMLUtils.findChildGetItsText(node, TAG_CONTENT));
		settings.setSaveToFile(
				TestExec.convertStateBoolean(
						XMLUtils.findChildGetItsText(node, TAG_SAVE_TO_FILE), false));
		settings.setDestFile(XMLUtils.findChildGetItsText(node, TAG_LOCAL_FILENAME));
		settings.setAuth(XMLUtils.findChildGetItsText(node, TAG_AUTH));
		settings.setPrivateKey(XMLUtils.findChildGetItsText(node, TAG_PRIVATE_KEY));

	}
	

    public void writeSubXML( PrintWriter pw )
    {
    	XMLUtils.streamTagAndChild(pw, TAG_HOSTNAME, settings.getHostname());
    	XMLUtils.streamTagAndChild(pw, TAG_PORT, settings.getPort());
    	XMLUtils.streamTagAndChild(pw, TAG_USERNAME, settings.getUsername());
    	
    	String tmp = StrUtil.encodeString(settings.getPassword());
    	XMLUtils.streamTagAndChild(pw, TAG_PASSWORD, tmp);
    	
    	String tmp2 = StrUtil.encodeString(settings.getPrivateKeyPass());
    	XMLUtils.streamTagAndChild(pw, TAG_PRIVATE_KEY_PASS, tmp2);

        
    	XMLUtils.streamTagAndChild(pw, TAG_PROTOCOL, settings.getProtocol());
    	XMLUtils.streamTagAndChild(pw, TAG_COMMAND, settings.getCommand());
    	XMLUtils.streamTagAndChild(pw, TAG_PATH, settings.getPath());
    	XMLUtils.streamTagAndChild(pw, TAG_FILENAME, settings.getFilename());
    	XMLUtils.streamTagAndChild(pw, TAG_CONTENT, settings.getContent());
    	XMLUtils.streamTagAndChild(pw, TAG_SAVE_TO_FILE, String.valueOf(settings.isSaveToFile()));
    	XMLUtils.streamTagAndChild(pw, TAG_LOCAL_FILENAME, settings.getDestFile());
    	XMLUtils.streamTagAndChild(pw, TAG_AUTH, settings.getAuth());
    	XMLUtils.streamTagAndChild(pw, TAG_PRIVATE_KEY, settings.getPrivateKey());
    }

	public String getTypeName() throws Exception {
		return "SFTP Step";
	}
}
