package com.itko.lisa.ext.trilead;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.pool.*;
import org.apache.commons.pool.impl.GenericObjectPool;

import com.itko.lisa.test.TestRunException;
import com.trilead.ssh2.*;

public class TrileadConnectionFactory implements PoolableObjectFactory, InteractiveCallback {

    private static Log log = LogFactory.getLog("com.itko.lisa.ext.ssh");
	private static Map<String,GenericObjectPool> poolMap = new HashMap<String,GenericObjectPool>();
	
	public static final String STEP_VERSION = "1.2";
	public static final String AUTH_PASSWORD = "Authenticate with Password";
	public static final String AUTH_KEYBOARD = "Authenticate with Keyboard Interactive";
	public static final String AUTH_PUBLIC_KEY = "Authenticate with Public Key";
	public static final String[] AUTH_TYPES = {
		AUTH_PASSWORD,
		AUTH_KEYBOARD,
		AUTH_PUBLIC_KEY
	};

    private String hostname;
    private int port;
    private String username;
    private String password;
    private String auth;
    private String privateKey;
    private String privateKeyPass;
	
	public TrileadConnectionFactory(String hostname,
			int port,
			String username,
			String password,
			String auth,
			String privateKey,
			String privateKeyPass) {
		
		this.hostname = hostname;
		this.port = port;
		this.username = username;
		this.password = password;
		this.auth = auth;
		this.privateKey = privateKey;
		this.privateKeyPass = privateKeyPass;
	}

	/**
	 * 
	 * @param settings
	 * @return
	 * @throws IOException
	 */
	protected static synchronized Connection getConnection(
			String tsHostname,
			int tsPort,
			String tsUsername,
			String tsPassword,
			String tsAuth,
			String tsPrivateKey,
			String tsPrivateKeyPass) throws Exception {
		
		// Try to find the client pool
		String key = tsUsername + "@" + tsHostname + ":" + tsAuth;
		GenericObjectPool clientPool = poolMap.get(key);
		if (clientPool == null) {
			
			// Create a new factory and pool, then add to the hashmaps
			log.debug("Creating a new() GenericObjectPool and Factory");
			
			// Create a new factory for these credentials
			TrileadConnectionFactory factory =
				new TrileadConnectionFactory(tsHostname, tsPort, tsUsername, tsPassword, tsAuth, tsPrivateKey, tsPrivateKeyPass);
			
			// Create a new pool & put into the hashmap
			// Size=5, block when empty, no timeout, test on borrow
			clientPool = new GenericObjectPool(
					factory,
					5,
					GenericObjectPool.WHEN_EXHAUSTED_BLOCK,
					0,
					true,
					false);
			poolMap.put(key, clientPool);
		}
		
		// Obtain a new connection
		Connection conn = null;
		try {
			conn = (Connection)clientPool.borrowObject();
		} catch (Exception e) {
			log.debug("***** Unable to borrow object");
			
			// Remove this entire pool from the HashMap
			poolMap.remove(key);
			throw new TestRunException(e.getMessage(), e);
		}
		return conn;
	}
	
	/**
	 * 
	 * @param conn
	 * @param tsHostname
	 * @param tsUsername
	 * @param invalidate
	 */
	protected static void cleanupConnection(Connection conn, String tsHostname, String tsUsername, String tsAuth, boolean invalidate) {
		try {
			String key = tsUsername + "@" + tsHostname + ":" + tsAuth;
			ObjectPool clientPool = poolMap.get(key);
			if (invalidate) {
				log.debug("Invalidating connection for: " + key);
				clientPool.invalidateObject(conn);
			} else {
				clientPool.returnObject(conn);
			}
			
		} catch (Exception e) {
			log.debug(e.getMessage(), e);
		}
	}
		
	/**
	 * 
	 * @return
	 * @throws IOException
	 */
	private Connection createNewConnection() throws Exception {
		log.debug("Creating new() Connection for " + username + "@" + hostname + " using " + auth + " authentication scheme.");
		
		// Create the SSL connection
		Connection conn = new Connection(hostname, port);
		
		// Change the connection
		//[aes256-ctr], [aes192-ctr], [aes128-ctr], [blowfish-ctr], [aes256-cbc], [aes192-cbc], [aes128-cbc], [blowfish-cbc], [3des-ctr], [3des-cbc]
		String[] cipherArr = new String[2];
		cipherArr[0] = "blowfish-ctr";
		cipherArr[1] = "blowfish-cbc";
		conn.setClient2ServerCiphers(cipherArr);
		conn.setServer2ClientCiphers(cipherArr);
		
		String[] cipers = Connection.getAvailableCiphers();
		String clist = "";
		for (int i=0; i<cipers.length; i++) {
			clist += "[" + cipers[i] + "], ";
		}
		log.debug("Entire list of ciphers available: " + clist);
		
		// Connect to the server
		conn.connect();
		ConnectionInfo cInfo = conn.getConnectionInfo();
		log.debug("Client crypto algorithm: " + cInfo.clientToServerCryptoAlgorithm);
		log.debug("Server crypto algorithm: " + cInfo.serverToClientCryptoAlgorithm);

		try {
			
			boolean isAuthenticated = checkAuth(conn, username);
			
			// Figure out if the authentication worked
			if (isAuthenticated == false) {
				String[] remainingAuths = conn.getRemainingAuthMethods(username);
				StringBuffer sbAuths = new StringBuffer();
				for (int i = 0; i < remainingAuths.length; i++) {
					sbAuths.append(remainingAuths[i] + "  ");
				}
				throw new IOException("Authentication failed for user: " + username + " but the other authentication schemes remain: " + sbAuths.toString());
			}
			
		} catch (IOException ioe) {
			log.debug("***** Unable to createNewConnection: " + ioe.getCause());
			throw new TestRunException(ioe.getMessage(), ioe);
		}
		
		return conn;
	}
	
	private boolean checkAuth( Connection conn, String username ) throws IOException {
		
		// Determine the authentication approach
		boolean isAuthenticated = false;
		if (auth.equalsIgnoreCase(AUTH_PASSWORD)) {
			// Authenticate with password
			isAuthenticated = conn.authenticateWithPassword(username, password);
		} else if (auth.equalsIgnoreCase(AUTH_KEYBOARD)) {
			// Authenticate with keyboard-interactive
			log.debug("authenticateWithKeyboardInteractive start");
			isAuthenticated = conn.authenticateWithKeyboardInteractive(username, this);
			log.debug("authenticateWithKeyboardInteractive finish");
		} else if (auth.equalsIgnoreCase(AUTH_PUBLIC_KEY)) {
			// Authenticate with private key
			isAuthenticated = conn.authenticateWithPublicKey(username, privateKey.toCharArray(), privateKeyPass);
		} else {
			log.error("*** Unknown Auth -- This shouldn't ever happen ***");
		}
		return isAuthenticated;
	}
	
	/**
	 * This is used by the keyboard-interactive approach 
	 * 
	 * @param name
	 * @param instruction
	 * @param numPrompts
	 * @param prompt
	 * @param echo
	 * @return
	 * @throws java.lang.Exception
	 */
	public String[] replyToChallenge(java.lang.String name,
            String instruction,
            int numPrompts,
            String[] prompt,
            boolean[] echo)
            throws java.lang.Exception {

		log.debug("replyToChallenge: start");
		
		String asteriskPW = new String("");
		for(int j=0; j < password.length(); j++) {
			asteriskPW += "*";
		}

		// Figure out the correct response to each challenge
		String[] reply = new String[numPrompts];
		for (int i = 0; i < numPrompts; i++) {
			log.debug("replyToChallenge: prompt[" + i + "]=" + prompt[i]);
			
			String sPrompt = prompt[i].toLowerCase().trim();
			
//			if (prompt[i].equalsIgnoreCase("Password: ")) {
			if (sPrompt.contains("password")) {
				log.debug("This prompt matches the password request: " + sPrompt);
				reply[i] = new String(password);
				log.debug("We put the following in the password field: " + asteriskPW + " len= " + password.length());
			} else {
				reply[i] = new String("");
			}
		}
		
		// Some debug logging
		log.debug("replyToChallenge: instruction = " + instruction);
		log.debug("replyToChallenge: numPrompts = " + numPrompts);
		for (int i = 0; i < prompt.length; i++) {
			log.debug("replyToChallenge: prompt[" + i + "] = " + prompt[i]);
		}
		for (int i = 0; i < reply.length; i++) {
			String asteriskReply = new String("");
			for(int j=0; j < reply[i].length(); j++) {
				asteriskReply += "*";
			}
			log.debug("replyToChallenge: reply[" + i + "] = " + asteriskReply + " len=" + reply[i].length());
		}
		for (int i = 0; i < echo.length; i++) {
			log.debug("replyToChallenge: echo[" + i + "] = " + echo[i]);
		}
		
		return reply;
	}
	
	public void activateObject(Object arg0) throws Exception {
	}

	public void destroyObject(Object arg0) throws Exception {
		((Connection)arg0).close();
	}

	public Object makeObject() throws Exception {
		Connection conn = createNewConnection();
		return conn;
	}

	public void passivateObject(Object arg0) throws Exception {
	}

	public boolean validateObject(Object arg0) {
		return testConnection((Connection)arg0);
	}
	
	public static boolean testConnection(Connection conn) {
		try {
			
			// If the ping() doesn't throw an exception we're good
			if (conn == null) {
				log.warn("The connection object could not be tested because it was null");
				return false;
			}
			conn.ping();
			return true;
			
		} catch (IOException ioe) {}
		
		log.warn("We failed to validate the SSH Connection!");
		return false;
	}
}
